/// \file ijkorient.cxx
// reorder isotable simplex vertices for consistent orientation
/// Version 0.2.2

/*
  IJK: Isosurface Jeneration Code
  Copyright (C) 2007, 2006 Rephael Wenger

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  (LGPL) as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <algorithm>
#include <fstream>
#include <string>
#include <vector>

#include <CGAL/MP_Float.h>
#include <CGAL/Lazy_exact_nt.h>
#include <CGAL/Cartesian_d.h>
#include <CGAL/Convex_hull_d.h>
#include <CGAL/enum.h>
#include <CGAL/predicates_d.h>

#include "ijk.txx"
#include "ijktable.h"
#include "ijkxitIO.h"

using namespace IJK;
using namespace std;

typedef CGAL::Cartesian_d < CGAL::Lazy_exact_nt < CGAL::Quotient <CGAL::MP_Float> > >    R;
typedef CGAL::Point_d<R>             Point;

// global variables
char * fname;
bool verbose = true;
int output_trigger = 5000;
u_int base = 2;
const u_int BINARY_NEGATIVE = 0;
const u_int BINARY_POSITIVE = 1;
const u_int BASE3_NEGATIVE = 0;
const u_int BASE3_POSITIVE = 2;
u_int negative = BINARY_NEGATIVE;
u_int positive = BINARY_POSITIVE;

// routines
void orient_isopatch
(IJKTABLE::ISOSURFACE_TABLE & isotable, const int it, float * coord);
void orient_volume
(IJKTABLE::ISOSURFACE_TABLE & isotable, const int it, float * coord);
int find_vertex_with_value
(const IJKTABLE::ISOSURFACE_TABLE & isotable, const int it, const int ival);
void copy_vertex_coord
(const IJKTABLE::ISOSURFACE_TABLE & isotable, const int iv, float * coord);
void flip_orientation
(IJKTABLE::ISOSURFACE_TABLE & isotable, const int it, const int is);
bool share_facet
(IJKTABLE::ISOSURFACE_TABLE & isotable, const int it, 
 const int is0, const int is1, int & f0, int & f1);
bool facet_orientation
(IJKTABLE::ISOSURFACE_TABLE & isotable, const int it,
 const int is, const int f);
void backup_file(const char * fname);
void parse_command_line(int argc, char **argv);
void memory_exhaustion();
void usage_error(), help();

int main(int argc, char **argv) {

  try {

    std::set_new_handler(memory_exhaustion);
    parse_command_line(argc, argv);

    IJKTABLE::ISOSURFACE_TABLE isotable(3);
    std::ifstream isotable_file(fname);

    if (!isotable_file) {
      string error_msg = 
	"Unable to open isosurface table file " + string(fname) + ".";
      throw(ERROR(error_msg));
    };

    try {
      IJKXIO::read_xit(isotable_file, isotable);
    }
    catch(...) {
      cerr << "Error reading file " << fname << "." << endl;
      throw;
    };

    isotable_file.close();

    if (isotable.Dimension() < 2) {
      cout << "Isotable dimension is " << isotable.Dimension() << "." << endl;
      cout << "No change to isosurface table." << endl;
      exit(0);
    };

    // compute base, positive and negative
    if (isotable.Encoding() == IJKTABLE::ISOSURFACE_TABLE::BINARY) {
      base = 2;
      negative = BINARY_NEGATIVE;
      positive = BINARY_POSITIVE;
    }
    else if (isotable.Encoding() == IJKTABLE::ISOSURFACE_TABLE::BASE3) {
      base = 3;
      negative = BASE3_NEGATIVE;
      positive = BASE3_POSITIVE;
    }
    else {
      std::string msg = 
	"Cannot orient isosurface table with encoding ";
      msg = msg + isotable.EncodingName() + ".";
      throw PROCEDURE_ERROR("main", msg);
    }

    // check that number of table entries is correct for given encoding 
    int num_vert = isotable.Polyhedron().NumVertices();
    u_long num_entries = isotable.NumTableEntries();
    u_long expected_num_entries = 0;
    expected_num_entries = IJKTABLE::calculate_num_entries(num_vert, base);


    if (num_entries != expected_num_entries) {
      ERROR error;
      string msg = string("Mismatch between isosurface table and ") +
	isotable.EncodingName() + " encoding.";
      error.AddMessage(msg);
      error.AddMessage("  Expected number of table entries = ", 
		       expected_num_entries, ".");
      msg = string("  Number of table entries in file ") +
	fname + " = ";
      error.AddMessage(msg, num_entries, ".");
      throw error;
    };

    assert(isotable.NumVerticesPerSimplex() > 1);

    float * coord = new float[isotable.Dimension()];

    if (isotable.Dimension() == isotable.SimplexDimension()+1) {
      for (int it = 0; it < isotable.NumTableEntries(); it++) {
	orient_isopatch(isotable, it, coord);

	if (verbose && it > 0 && it%output_trigger == 0) {
	  cout << "  " << it << " out of " << isotable.NumTableEntries()
	       << " isosurface table entries completed." << endl; 
	};
      };
    }
    else if (isotable.Dimension() == isotable.SimplexDimension()) {
      if (base != 3) {
	ERROR error;
	error.AddMessage("Illegal base ", base, ".");
	error.AddMessage("Base must equal 3 when simplex dimension equals volume dimension.");
	throw error;
      };

      for (int it = 0; it < isotable.NumTableEntries(); it++) {
	orient_volume(isotable, it, coord);

	if (verbose && it > 0 && it%output_trigger == 0) {
	  cout << "  " << it << " out of " << isotable.NumTableEntries()
	       << " isosurface table entries completed." << endl; 
	};
      };
    }
    else {
      int d = isotable.Dimension();
      int d2 = isotable.SimplexDimension();
      ERROR error;
      error.AddMessage
	("Orientation not implemented for dimension ", d, 
	 ", simplex dimension ", d2, ".");
      error.AddMessage("Dimension and simplex dimension must not differ by more than one.");
      throw(error);
    };

    if (verbose && isotable.NumTableEntries() > output_trigger) {
      cout << "  All " << isotable.NumTableEntries()
	   << " isosurface table entries completed." << endl; 
    }

    delete [] coord;

    backup_file(fname);

    std::ofstream isotable_file_out(fname);


    try {
      IJKXIO::write_xit(isotable_file_out, isotable);
    }
    catch(...) {
      cerr << "Error writing file " << fname
	   << "." << endl;
      throw;
    }
  }
  catch (ERROR error) {
    if (error.NumMessages() == 0) {
      cerr << "Unknown error." << endl;
    }
    else { error.Print(cerr); }
    cerr << "Exiting." << endl;
    exit(20);
  }
  catch (...) {
    cerr << "Unknown error." << endl;
    exit(50);
  };

  exit(0);
}

void orient_isopatch
(IJKTABLE::ISOSURFACE_TABLE & isotable, const int it, float * coord)
  // orient isosurface patch for table entry it
  // isotable = isosurface table
  // it = index of table entry
  // coord[] = array for storing coordinates
  //   Precondition: coord[] array must be preallocated
  //   with size at least isotable.Dimension()
{
  PROCEDURE_ERROR error("orient_isopatch");

  bool * orient_flag = NULL;
  int * digits = NULL;

  try {

    int nums = isotable.NumSimplices(it);
    int nume = isotable.Polyhedron().NumEdges();
    int numv = isotable.Polyhedron().NumVertices();
    if (nums <= 0) return;

    digits = new int[numv];
    IJKTABLE::convert2base(it, base, digits, numv);

    
    int iv_positive = find_vertex_with_value(isotable, it, positive);
    int iv_negative = find_vertex_with_value(isotable, it, negative);

    bool is_q_positive = true;
    if (iv_positive < isotable.Polyhedron().NumVertices()) {
      copy_vertex_coord(isotable, iv_positive, coord);
    }
    else if (iv_negative < isotable.Polyhedron().NumVertices()) {
      copy_vertex_coord(isotable, iv_negative, coord);
      is_q_positive = false;
    }
    else {
      cerr << "No positive and no negative vertices at table entry " 
	   << it << "."
	   << "  Skipping entry." << endl;
      return;
    };

    Point q0(isotable.Dimension(), coord, coord+isotable.Dimension());

    bool * orient_flag = new bool[nums];
    for (int i = 0; i < nums; i++)
      orient_flag[i] = false;

    for (int is = 0; is < nums; is++) {

      vector<Point> p;

      bool skip_simplex = false;
      for (int jv = 0; jv < isotable.NumVerticesPerSimplex(); jv++) {
	int ie = isotable.SimplexVertex(it, is, jv);
	
	if (base == 2) {
	  if (ie < nume) {
	    for (int ic = 0; ic < isotable.Dimension(); ic++)
	      coord[ic] = isotable.Polyhedron().MidpointCoord(ie, ic);
	  }
	  else {
	    // can't determine the coordinates represented by this vertex
	    skip_simplex = true;
	  }

	}
	else if (base == 3) {
	  if (ie < numv) {
	    int iv = ie;
	    copy_vertex_coord(isotable, iv, coord);
	  }
	  else if (ie-numv < nume) {
	    int ie2 = ie-numv;

	    for (int ic = 0; ic < isotable.Dimension(); ic++)
	      coord[ic] = isotable.Polyhedron().MidpointCoord(ie2, ic);
	  }
	  else {
	    // can't determine the coordinates represented by this vertex
	    skip_simplex = true;
	  }
	}
	else {
	  error.AddMessage("Illegal base.");
	  throw error;
	};

	if (!skip_simplex) {
	  Point q(isotable.Dimension(), coord, coord+isotable.Dimension());
	  p.push_back(q);
	}
      };

      if (!skip_simplex) {

	for (int iv = 0; iv < numv; iv++)
	  if (digits[iv] != negative) {
	    // iv is either positive or zero
	    copy_vertex_coord(isotable, iv, coord);

	    Point q(isotable.Dimension(), coord, coord+isotable.Dimension());
	    p.push_back(q);

	    CGAL::Orientation orient = CGAL::orientation(p.begin(), p.end());

	    if (orient != CGAL::ZERO) {
	      if (orient == CGAL::NEGATIVE) {
		flip_orientation(isotable, it, is);
	      };
	      orient_flag[is] = true;
	      break;
	    };

	    p.pop_back();
	  }
      }

      if (!orient_flag[is] && !skip_simplex) {
	// simplex lies in some polyhedron facet

	for (int iv = 0; iv < numv; iv++)
	  if (digits[iv] == negative) {
	    // iv is negative
	    copy_vertex_coord(isotable, iv, coord);

	    Point q(isotable.Dimension(), coord, coord+isotable.Dimension());
	    p.push_back(q);

	    CGAL::Orientation orient = CGAL::orientation(p.begin(), p.end());

	    if (orient != CGAL::ZERO) {
	      if (orient == CGAL::POSITIVE) {
		flip_orientation(isotable, it, is);
	      };
	      orient_flag[is] = true;
	      break;
	    };

	    p.pop_back();
	  }
      }
    };

    for (int is = 0; is < nums; is++) {
      if (!orient_flag[is]) {
	cerr << "Unable to orient table entry " << it
	     << " simplex " << is << "." << endl;
      };
    };

  }
  catch(...) {
    delete [] orient_flag;
    orient_flag = NULL;
    delete [] digits;
    digits = NULL;

    throw;
  };

  delete [] orient_flag;
  orient_flag = NULL;
  delete [] digits;
  digits = NULL;
}

void orient_volume
(IJKTABLE::ISOSURFACE_TABLE & isotable, const int it, float * coord)
  // orient volume for table entry it
  // isotable = isosurface table
  // it = index of table entry
  // coord[] = array for storing coordinates
  //   Precondition: coord[] array must be preallocated
  //   with size at least isotable.Dimension()
{
  PROCEDURE_ERROR error("orient_volume");
  int * digits = NULL;

  assert(base == 3);

  typedef IJKTABLE::ISOSURFACE_VERTEX::ISOSURFACE_VERTEX_TYPE
    ISOSURFACE_VERTEX_TYPE;
  const ISOSURFACE_VERTEX_TYPE VERTEX = IJKTABLE::ISOSURFACE_VERTEX::VERTEX;
  const ISOSURFACE_VERTEX_TYPE EDGE = IJKTABLE::ISOSURFACE_VERTEX::EDGE;

  string error_msg = "";

  try {

    int nume = isotable.Polyhedron().NumEdges();
    int numv = isotable.Polyhedron().NumVertices();
    int nums = isotable.NumSimplices(it);
    if (nums <= 0) return;

    digits = new int[numv];
    IJKTABLE::convert2base(it, base, digits, numv);

    for (int is = 0; is < nums; is++) {

      vector<Point> p;

      error_msg = "";
      bool skip_simplex = false;
      for (int jv = 0; jv < isotable.NumVerticesPerSimplex(); jv++) {
	int kv = isotable.SimplexVertex(it, is, jv);
	const IJKTABLE::ISOSURFACE_VERTEX isov = 
	  isotable.IsosurfaceVertex(kv);
	ISOSURFACE_VERTEX_TYPE vtype = isov.Type();
	
	if (vtype == EDGE) {
	  int ie = isov.Face();
	  int iv0 = isotable.Polyhedron().EdgeEndpoint(ie, 0);
	  int iv1 = isotable.Polyhedron().EdgeEndpoint(ie, 1);
	  if (digits[iv0] > digits[iv1]) {
	    swap(iv0, iv1);
	  }

	  // all coordinates are implicitly multiplied by 3
	  // so that coord[] contains integers
	  if (isov.IsLabelSet()) {
	    if (isov.Label() == "0") {
	      for (int ic = 0; ic < isotable.Dimension(); ic++) {
		int c0 = isotable.Polyhedron().VertexCoord(iv0, ic);
		int c1 = isotable.Polyhedron().VertexCoord(iv1, ic);
		coord[ic] = 2*c0 + c1;
	      }
	    }
	    else if (isov.Label() == "1") {
	      for (int ic = 0; ic < isotable.Dimension(); ic++) {
		int c0 = isotable.Polyhedron().VertexCoord(iv0, ic);
		int c1 = isotable.Polyhedron().VertexCoord(iv1, ic);
		coord[ic] = c0 + 2*c1;
	      }
	    }
	    else {
	      skip_simplex = true;
	      error_msg = "Unknown label " + isov.Label() + ".";
	    }
	  }
	  else {
	    skip_simplex = true;
	    error_msg = "Label not set.";
	  };
	}
	else if (vtype == VERTEX) {
	  int iv = isov.Face();
	  for (int ic = 0; ic < isotable.Dimension(); ic++)
	    coord[ic] = 3*isotable.Polyhedron().VertexCoord(iv, ic);
	}
	else {
	  skip_simplex = true;
	  error_msg = "Illegal vertex type.";
	};

	if (!skip_simplex) {
	  Point q(isotable.Dimension(), coord, coord+isotable.Dimension());
	  p.push_back(q);
	}
      };

      if (!skip_simplex) {

	CGAL::Orientation orient = CGAL::orientation(p.begin(), p.end());

	switch (orient) {

	case CGAL::NEGATIVE:
	  flip_orientation(isotable, it, is);
	  break;

	case CGAL::POSITIVE:
	  break;

	case CGAL::ZERO:
	  error_msg = "Degenerate simplex.";
	  skip_simplex = true;
	  break;
	};
      };

      if (skip_simplex) {
	cerr << "Unable to orient table entry " << it
	     << " simplex " << is << ".  " << error_msg << endl;
      }
    };

  }
  catch(...) {
    delete [] digits;
    digits = NULL;

    throw;
  };

  delete [] digits;
  digits = NULL;
}

int find_vertex_with_value
(const IJKTABLE::ISOSURFACE_TABLE & isotable, const int it, const int ival)
  // return a vertex with value ival
  // if there no vertex with that value, return a value >= num_vertices
{
  int numv = isotable.Polyhedron().NumVertices();
  int * digits = new int[numv];
  int iv= 0;

  try {
    IJKTABLE::convert2base(it, base, digits, numv);

    while (iv < numv && digits[iv] != ival) 
      { iv++; };
  }
  catch(...) {
    delete [] digits;
    digits = NULL;
    throw;
  };

  delete [] digits;
  digits = NULL;

  return(iv);
}

void copy_vertex_coord
(const IJKTABLE::ISOSURFACE_TABLE & isotable, const int iv, float * coord)
  // copy vertex coordinate into array coord[]
  // isotable = isosurface lookup table
  // iv = polyhedron vertex
  // coord[] = coordinate array
  //   Precondition: coord[] array must be preallocated
  //   with size at least isotable.Dimension()
{
  assert(coord != NULL);
  assert(iv < isotable.Polyhedron().NumVertices());

  for (int ic = 0; ic < isotable.Dimension(); ic++)
    coord[ic] = isotable.Polyhedron().VertexCoord(iv, ic);
}


void flip_orientation
(IJKTABLE::ISOSURFACE_TABLE & isotable, const int it, const int is)
  // flip the orientation of simplex is in table entry it
  // by swapping the last two vertices of the simplex
{
  // swap last two vertices
  int j1 = isotable.NumVerticesPerSimplex()-1;
  int j0 = j1 - 1;
  int je0 = isotable.SimplexVertex(it, is, j0);
  int je1 = isotable.SimplexVertex(it, is, j1);
  isotable.SetSimplexVertex(it, is, j0, je1);
  isotable.SetSimplexVertex(it, is, j1, je0);
}

bool share_facet
(IJKTABLE::ISOSURFACE_TABLE & isotable, const int it, 
 const int is0, const int is1, int & f0, int & f1)
{
  bool * vmatch0 = NULL;
  bool * vmatch1 = NULL;

  f0 = f1 = 0;                 // initialize

  int numv_per_simplex = isotable.NumVerticesPerSimplex();
  vmatch0 = new bool[numv_per_simplex];
  vmatch1 = new bool[numv_per_simplex];

  for (int iv = 0; iv < numv_per_simplex; iv++) {
    vmatch0[iv] = false;
    vmatch1[iv] = false;
  };

  for (int iv0 = 0; iv0 < numv_per_simplex; iv0++) {
    for (int iv1 = 0; iv1 < numv_per_simplex; iv1++) {
      int ie0 = isotable.SimplexVertex(it, is0, iv0);
      int ie1 = isotable.SimplexVertex(it, is1, iv1);
      if (ie0 == ie1) {
	vmatch0[iv0] = true;
	vmatch1[iv1] = true;
      };
    };
  };

  int num_unmatched0 = 0;
  int jf0 = 0;
  for (int iv0 = 0; iv0 < numv_per_simplex; iv0++) {
    if (!vmatch0[iv0]) {  
      num_unmatched0++; 
      jf0 = iv0;
    };
  };

  int num_unmatched1 = 0;
  int jf1 = 0;
  for (int iv1 = 0; iv1 < numv_per_simplex; iv1++) {
    if (!vmatch1[iv1]) {  
      num_unmatched1++; 
      jf1 = iv1;
    };
  };

  delete [] vmatch0;
  delete [] vmatch1;

  if (num_unmatched0 == 1 && num_unmatched1 == 1) {
    f0 = jf0;
    f1 = jf1;
    return(true);
  };

  return(false);
}

bool facet_orientation
(IJKTABLE::ISOSURFACE_TABLE & isotable, const int it,
 const int is, const int f)
  // f = facet index.  All vertices are on facet except vertex f.
{
  bool orientation = true;
  int * vlist = NULL;
  int numv_per_simplex = isotable.NumVerticesPerSimplex();

  assert(f < numv_per_simplex);

  vlist = new int[numv_per_simplex-1];

  int k = 0;
  for (int iv = 0; iv < numv_per_simplex; iv++) {
    if (iv != f) {
      int jv = isotable.SimplexVertex(it, is, iv);
      vlist[k] = jv;
      k++;
    };
  };

  if (f % 2 == 0) { orientation = true; }
  else { orientation = false; };

  bool done = false;
  while (!done) {
    done = true;

    for (int iv = 0; iv < numv_per_simplex-1; iv++) {
      if (vlist[iv] > vlist[iv+1]) {
	swap(vlist[iv], vlist[iv+1]);
	orientation = (!orientation);
      }
    };
  };

  delete [] vlist;

  return(orientation);
}

void backup_file(const char * fname)
{
  std::string command;
  command = "mv " + std::string(fname) + " " + std::string(fname) + "~";
  system(command.c_str());
}

void parse_command_line(int argc, char **argv)
{
  int iarg = 1;
  while (iarg < argc) {
    if (argv[iarg][0] != '-')
      break;

    if (strcmp(argv[iarg], "-help") == 0) {
      help();
    }
    else {
      for (int j = 1; argv[iarg][j] != '\0'; j++) {
	if (argv[iarg][j] == 'h')
	  help();
	else if (argv[iarg][j] == 'q')
	  verbose = false;
	else if (argv[iarg][j] == 'v')
	  verbose = true;
	else
	  usage_error();
      };
    };

    iarg++;
  };
  
  // get file name
  if (iarg+1 != argc) { usage_error(); };
  fname = argv[iarg];
}

void memory_exhaustion()
{
   cerr << "Error: Out of memory.  Terminating program." << endl;
   exit(10);
}

void usage_msg()
{
  cerr << "Usage: ijkorient -help {isotable file}" << endl;
}

void usage_error()
{
  usage_msg();
  exit(10);
}

void help()
{
  usage_msg();
  cerr << endl;
  cerr << "ijkorient - Orient isosurface table." << endl;
  cerr << endl;
  cerr << "Options:" << endl;
  cerr << "  -help: Print this help message." << endl;
  exit(20);
}
